%% This code is an adaptation of the Cobra Toolbox OptKnock Tutorial developed by Sebastian N. Martinez
% source: https://opencobra.github.io/cobratoolbox/stable/tutorials/tutorialOptKnock.html
% Adapted by: Victor A. Lopez-Agudelo, University of Valle - Cali - Colombia
clc; clear;

% SolverOK = changeCobraSolver('ibm_cplex','all');

modelFileName = 'sclav_red_fedbatch.mat'; % sclav_red constrained at 48 h of fed-batch stage
model = readCbModel(modelFileName);
biomass = 'Growth';

% Define the maximum number of solutions to find 
% (i.e., maximum number of remvable reactions that lead to the overproduction of the metabolite of interest)
threshold = 5;

% Define the set of reactions that will be used to search knockouts. 
% Note : only reactions in this set will be deleted

selectedRxnList  = {'PGI';'PGM';'G3PD1';'G6PDH2r';'GAPD';'PGK';'PYK';'PPC';'PPCK';'ME1';'MDH';'SUCD1';'ICDHyr';
				 'ICL';'ASPTA';'RPI';'TKT1';'TKT2';'TALA';'CBPS';'CMC1';'CS1';'CS2';'CS3';'CEAS'; 'PFK'; 'FBA';...
				 'TPI'; 'PGM'; 'ENO'; 'LDH_D'; 'L_LACD2'; 'PFL'; 'ALCD2x'; 'PTAr'; 'ACKr'; 'PGL'; 'GND'; 'RPI'; 
				 'RPE'; 'FUM'; 'SUCOAS'; 'AKGDH'; 'ACONT';'CS'; 'MDH';'MDH2'; 'MDH3'; 'ACALD';'ME2';'GLUDxi';...
                 };
				 
%% Relaxing constraints of growth rate and set it as default objective function

% set the default objective function to be Growth Rate
model = changeObjective(model,'Growth');

% defining medium for sclav
medium_base = {'EX_h2o_e','EX_h_e','EX_cl_e','EX_nh4_e','EX_k_e','EX_ca2_e','EX_cit_e','EX_mg2_e','EX_mn2_e',...
                'EX_cobalt2_e','EX_zn2_e','EX_cu2_e','EX_fe2_e', 'EX_fe3_e','EX_mobd_e','EX_so4_e', 'EX_cbl1_e', 'EX_ni2_e'};

medium_trace = {'EX_glu__L_e'};


oxygen = 'EX_o2_e';
phosphate = 'EX_pi_e';
co2 = 'EX_co2_e';
glycerol = 'EX_glyc_e';
citrate = 'EX_cit_e';
clavulanate = 'EX_clav_e';

model = changeRxnBounds(model,medium_trace,-0.35,'l');
model = changeRxnBounds(model,medium_base,-1000,'l');
model = changeRxnBounds(model,'Growth',0.0, 'l'); % set lower bound growth rate
model = changeRxnBounds(model,'Growth',1000,'u'); % set upper bound growth rate
model = changeRxnBounds(model,oxygen,-1.2,'l');
model = changeRxnBounds(model,phosphate,-10,'l');
model = changeRxnBounds(model,phosphate,0,'u');
model = changeRxnBounds(model,glycerol,-0.470,'l'); 
model = changeRxnBounds(model,co2,1.4,'u'); 
model = changeRxnBounds(model,citrate,0,'l');
model = changeRxnBounds(model,clavulanate,0.0,'l'); % set lower bound to 0.0
model = changeRxnBounds(model,clavulanate,1000,'u'); % set upper bound to 1000
model = changeRxnBounds(model,'EX_succ_e',0.007,'l');
model = changeRxnBounds(model,'EX_oaa_e',0.003,'u');
model = changeRxnBounds(model,'EX_oaa_e',1000,'u');
model = changeRxnBounds(model,'EX_mal__L_e',0,'u');
model = changeRxnBounds(model,'EX_pyr_e',0.005,'l');
model = changeRxnBounds(model,'EX_pyr_e',1000,'u');
model = changeRxnBounds(model,'EX_ac_e',0,'u');

%model = changeRxnBounds(model,'ATPM',3.02,'b'); % NGAM

%% determine succinate production and growth rate
fbaWT = optimizeCbModel(model,'max');
clavFluxWT = fbaWT.x(strcmp(model.rxns, 'EX_clav_e'));
succFluxWT = fbaWT.x(strcmp(model.rxns, 'EX_succ_e'));
%etohFluxWT = fbaWT.x(strcmp(model.rxns, 'EX_etoh_e'));
%formFluxWT = fbaWT.x(strcmp(model.rxns, 'EX_for_e'));
%lactFluxWT = fbaWT.x(strcmp(model.rxns, 'EX_lac__D_e'));
acetFluxWT = fbaWT.x(strcmp(model.rxns, 'EX_ac_e'));
growthRateWT = fbaWT.f;
fprintf('The production of clavulanic acid before optimization is %f \n', clavFluxWT);
fprintf('The growth rate before optimization is %f \n', growthRateWT);
fprintf(['The production of other products such as succinate, and'...
'acetate are %f, and %f, respectively. \n'], ...
succFluxWT, acetFluxWT);

%% Running OptKnock
fprintf('\n...Finding optKnock sets of size 3 or less...\n\n')

% Set optKnock options
% The exchange of clavulanic acid will be the objective of the outer problem
options = struct('targetRxn', 'EX_clav_e', 'numDel', 3);
% We will impose that biomass be at least 50% of the biomass of wild-type
constrOpt = struct('rxnList', {{biomass}}, 'values', 0.3*fbaWT.f, 'sense', 'G');
% We will try to find 10 optKnock sets of a maximun length of 3
previousSolutions = cell(10, 1);
contPreviousSolutions = 1;
nIter = 1;
while nIter < threshold
fprintf('...Performing optKnock analysis...')
if isempty(previousSolutions{1})
optKnockSol = OptKnock(model, selectedRxnList, options, constrOpt);
else
optKnockSol = OptKnock(model, selectedRxnList, options, constrOpt, previousSolutions);
end
% determine succinate production and growth rate after optimization
clavFluxM1 = optKnockSol.fluxes(strcmp(model.rxns, 'EX_clav_e'));
growthRateM1 = optKnockSol.fluxes(strcmp(model.rxns, biomass));
succFluxM1 = optKnockSol.fluxes(strcmp(model.rxns, 'EX_succ_e'));
acetFluxM1 = optKnockSol.fluxes(strcmp(model.rxns, 'EX_ac_e'));
setM1 = optKnockSol.rxnList;
if ~isempty(setM1)
previousSolutions{contPreviousSolutions} = setM1;
contPreviousSolutions=contPreviousSolutions + 1;
%printing results
fprintf('optKnock found a optKnock set of large %d composed by ',length(setM1));
for j = 1:length(setM1)
if j == 1
fprintf('%s',setM1{j});
elseif j == length(setM1)
fprintf(' and %s',setM1{j});
else
fprintf(', %s',setM1{j});
end
end
fprintf('\n');
fprintf('The production of clavulanic acid after optimization is %.4f \n', clavFluxM1);
fprintf('The growth rate after optimization is %.4f \n', growthRateM1);
fprintf(['The production of other products such as succinate and acetate are ' ...
'%.4f, and %.4f, respectively. \n'], succFluxM1, acetFluxM1);
fprintf('...Performing coupling analysis...\n');
[type, maxGrowth, maxProd, minProd] = analyzeOptKnock(model, setM1, 'EX_clav_e');
fprintf('The solution is of type: %s\n', type);
fprintf('The maximun growth rate given the optKnock set is %.42f\n', maxGrowth);
fprintf(['The maximun and minimun production of clavulanic acid given the optKnock set is ' ...
'%.4f and %.4f, respectively \n\n'], minProd, maxProd);
%if strcmp(type,'growth coupled')
singleProductionEnvelope(model, setM1, 'EX_clav_e', biomass, 'savePlot', 1, 'showPlot', 1, ...
'fileName', ['clav_ex2_' num2str(nIter)], 'outputFolder', 'OptKnockResults');
%end
else
if nIter == 1
fprintf('optKnock was not able to found an optKnock set\n');
else
fprintf('optKnock was not able to found additional optKnock sets\n');
end
break;
end
nIter = nIter + 1;
end